#include "Plane3.h"
#include "ViewCellBsp.h"
#include "Mesh.h"
#include "common.h"
#include "ViewCell.h"
#include "Environment.h"
#include "Polygon3.h"
#include "Ray.h"
#include "AxisAlignedBox3.h"
#include <stack>
#include <time.h>
#include <iomanip>
#include "Exporter.h"

int BspTree::sTermMaxPolygons = 10;
int BspTree::sTermMaxDepth = 20;
int BspTree::sMaxCandidates = 10;
int BspTree::sSplitPlaneStrategy = NEXT_POLYGON; 
int BspTree::sConstructionMethod = VIEW_CELLS;

/** Evaluates split plane classification with respect to the plane's 
	contribution for a balanced tree.
*/
float BspTree::sLeastSplitsTable[] = {0, 0, 1, 0};
/** Evaluates split plane classification with respect to the plane's 
	contribution for a minimum number splits in the tree.
*/
float BspTree::sBalancedTreeTable[] = {-1, 1, 0, 0};

//int counter = 0;

/****************************************************************/
/*                  class BspNode implementation                */
/****************************************************************/

BspNode::BspNode(): mParent(NULL), mPolygons(NULL)//,mViewCellIdx(0)
{}

BspNode::BspNode(BspInterior *parent): mParent(parent), mPolygons(NULL)//,mViewCellIdx(0)
{}

BspNode::~BspNode()
{
	if (mPolygons)
	{
		CLEAR_CONTAINER(*mPolygons);
		DEL_PTR(mPolygons);
	}
}

bool BspNode::IsRoot() const 
{
	return mParent == NULL;
}

BspInterior *BspNode::GetParent() 
{ 
	return mParent; 
}

void BspNode::SetParent(BspInterior *parent)
{
	mParent = parent;
}

PolygonContainer *BspNode::GetPolygons()
{
	if (!mPolygons)
		mPolygons = new PolygonContainer();

	return mPolygons;
}

void BspNode::ProcessPolygons(PolygonContainer *polys, bool storePolys)
{
	if (storePolys)
	{
		while (!polys->empty())
		{
			GetPolygons()->push_back(polys->back());
			polys->pop_back();
		}
	}
	else CLEAR_CONTAINER(*polys);
}


/****************************************************************/
/*              class BspInterior implementation                */
/****************************************************************/


BspInterior::BspInterior(const Plane3 &plane): 
mPlane(plane), mFront(NULL), mBack(NULL)
{}

bool BspInterior::IsLeaf() const
{ 
	return false; 
}

BspNode *BspInterior::GetBack() 
{
	return mBack;
}

BspNode *BspInterior::GetFront() 
{
	return mFront;
}

Plane3 *BspInterior::GetPlane()
{
	return &mPlane;
}

void BspInterior::ReplaceChildLink(BspNode *oldChild, BspNode *newChild) 
{
	if (mBack == oldChild)
		mBack = newChild;
	else
		mFront = newChild;
}

void BspInterior::SetupChildLinks(BspNode *b, BspNode *f) 
{
    mBack = b;
    mFront = f;
}

void BspInterior::ProcessPolygon(Polygon3 **poly, const bool storePolys)
{
	if (storePolys) 
		GetPolygons()->push_back(*poly);
	else
		DEL_PTR(*poly);
}

void BspInterior::SplitPolygons(PolygonContainer *polys, 
								PolygonContainer *frontPolys, 
								PolygonContainer *backPolys, 
								PolygonContainer *coincident,
								int &splits, 
								bool storePolys)
{
	Polygon3 *splitPoly = NULL;
#ifdef _Debug
	Debug << "splitting polygons of node " << this << " with plane " << mPlane << endl;
#endif
	while (!polys->empty())
	{
		Polygon3 *poly = polys->back();
		polys->pop_back();

		//Debug << "New polygon with plane: " << poly->GetSupportingPlane() << "\n";

		// get polygon classification
		int classification = poly->ClassifyPlane(mPlane);

		Polygon3 *front_piece = NULL;
		Polygon3 *back_piece = NULL;
	
		switch (classification)
		{
			case Polygon3::COINCIDENT:
				//Debug << "coincident" << endl;	
				coincident->push_back(poly);
				break;			
			case Polygon3::FRONT_SIDE:	
				//Debug << "front" << endl;
				frontPolys->push_back(poly);
				break;
			case Polygon3::BACK_SIDE:
				//Debug << "back" << endl;
				backPolys->push_back(poly);
				break;
			case Polygon3::SPLIT:
				front_piece = new Polygon3(poly->mParent);
				back_piece = new Polygon3(poly->mParent);

				//-- split polygon into front and back part
				poly->Split(&mPlane, front_piece, back_piece);

				++ splits; // increase number of splits
				// check if polygon still valid 
				if (front_piece->mVertices.size() >= 3)
					frontPolys->push_back(front_piece);
				else
					DEL_PTR(front_piece);
				// check if polygon still valid
				if (back_piece->mVertices.size() >= 3)
					backPolys->push_back(back_piece);
				else				
					DEL_PTR(back_piece);
				
#ifdef _DEBUG
				Debug << "split " << *poly << endl << *front_piece << endl << *back_piece << endl;
#endif
				ProcessPolygon(&poly, storePolys);
				
				break;
			default:
                Debug << "SHOULD NEVER COME HERE\n";
				break;
		}
	}
}

/****************************************************************/
/*                  class BspLeaf implementation                */
/****************************************************************/
BspLeaf::BspLeaf(): mViewCell(NULL)
{
}

BspLeaf::BspLeaf(ViewCell *viewCell): mViewCell(viewCell)
{
}

BspLeaf::BspLeaf(BspInterior *parent): BspNode(parent), mViewCell(NULL)
{}

BspLeaf::BspLeaf(BspInterior *parent, ViewCell *viewCell): 
BspNode(parent), mViewCell(viewCell)
{
}
ViewCell *BspLeaf::GetViewCell()
{
	return mViewCell;
}

void BspLeaf::SetViewCell(ViewCell *viewCell)
{
	mViewCell = viewCell;
}

bool BspLeaf::IsLeaf() const 
{ 
	return true; 
}

/****************************************************************/
/*                  class BspTree implementation                */
/****************************************************************/

BspTree::BspTree(): 
mRoot(NULL), 
mVerticalAxis(Vector3(0, 0, 1)),
//mStorePolys(true)
mStorePolys(false)
{
	Randomize(); // initialise random generator for heuristics
}
 
const BspTreeStatistics &BspTree::GetStatistics() const 
{
	return mStat;
}

void BspTreeStatistics::Print(ostream &app) const
{
	app << "===== BspTree statistics ===============\n";

	app << setprecision(4);

	app << "#N_RAYS Number of rays )\n"
		<< rays << endl;
	app << "#N_DOMAINS  ( Number of query domains )\n"
		<< queryDomains <<endl;

	app << "#N_NODES ( Number of nodes )\n" << nodes << "\n";

	app << "#N_INTERIORS ( Number of interior nodes )\n" << Interior() << "\n";

	app << "#N_LEAVES ( Number of leaves )\n" << Leaves() << "\n";

	app << "#N_SPLITS ( Number of splits )\n" << splits << "\n";

	app << "#N_RAYREFS ( Number of rayRefs )\n" <<
	rayRefs << "\n";

	app << "#N_RAYRAYREFS  ( Number of rayRefs / ray )\n" <<
	rayRefs/(double)rays << "\n";

	app << "#N_LEAFRAYREFS  ( Number of rayRefs / leaf )\n" <<
	rayRefs/(double)Leaves() << "\n";

	app << "#N_MAXOBJECTREFS  ( Max number of rayRefs / leaf )\n" <<
	maxObjectRefs << "\n";

	app << "#N_NONEMPTYRAYREFS  ( Number of rayRefs in nonEmpty leaves / non empty leaf )\n" <<
	rayRefsNonZeroQuery/(double)(Leaves() - zeroQueryNodes) << "\n";

	app << "#N_LEAFDOMAINREFS  ( Number of query domain Refs / leaf )\n" <<
	objectRefs/(double)Leaves() << "\n";

	app << "#N_PEMPTYLEAVES  ( Percentage of leaves with zero query domains )\n"<<
	zeroQueryNodes*100/(double)Leaves() << endl;

	app << "#N_PMAXDEPTHLEAVES ( Percentage of leaves at maxdepth )\n"<<
	maxDepthNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMAXDEPTH ( Maximal reached depth )\n" << maxDepth << endl;

	app << "#N_PMINDEPTH ( Minimal reached depth )\n" << minDepth << endl;

	app << "#AVGDEPTH ( average depth )\n" << AvgDepth() << endl;

	app << "#N_ADDED_RAYREFS  ( Number of dynamically added ray references )\n"<<
	addedRayRefs << endl;

	app << "#N_REMOVED_RAYREFS  ( Number of dynamically removed ray references )\n" <<
	removedRayRefs << endl;

	app << "#N_INPUT_POLYGONS (number of input polygons )\n" <<
		polys << endl;

	app << "#N_ROUTPUT_INPUT_POLYGONS ( ratio polygons after subdivision / input polygons )\n" <<
		 (polys + splits) / (double)polys << endl;
	
	app << "===== END OF BspTree statistics ==========\n";
}

BspTree::~BspTree()
{
	std::stack<BspNode *> tStack;

	tStack.push(mRoot);

	while (!tStack.empty())
	{
		BspNode *node = tStack.top();

	    tStack.pop();
	
		if (!node->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			// push the children on the stack (there are always two children)
			tStack.push(interior->GetBack());
			tStack.push(interior->GetFront());
		}

		DEL_PTR(node);
	}
}

void BspTree::InsertViewCell(ViewCell *viewCell)
{
	PolygonContainer *polys = new PolygonContainer();

	// extract polygons that guide the split process
	mStat.polys += AddMesh2Polygons(viewCell->GetMesh(), *polys);
	mBox.Include(viewCell->GetBox()); // add to BSP aabb

	InsertPolygons(polys);
}

void BspTree::InsertPolygons(PolygonContainer *polys, ViewCellContainer *viewCells)
{	
	std::stack<BspTraversalData> tStack;
		
	// traverse tree or create new one
	BspNode *firstNode = mRoot ? mRoot : new BspLeaf();

	tStack.push(BspTraversalData(firstNode, polys, 0, NULL));

	while (!tStack.empty())
	{
		// filter polygons donw the tree
		BspTraversalData tData = tStack.top();
	    tStack.pop();
			
		if (!tData.mNode->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(tData.mNode);

			//-- filter view cell polygons down the tree until a leaf is reached
			if ((int)tData.mPolygons->size() > 0)
			{
				PolygonContainer *frontPolys = new PolygonContainer();
				PolygonContainer *backPolys = new PolygonContainer();
				PolygonContainer coincident;

				int splits = 0;
		
				// split viecell polygons with respect to split plane
				interior->SplitPolygons(tData.mPolygons, 
										frontPolys, 
										backPolys,
										&coincident,
										splits, 
										mStorePolys);
				
				// get view cell associated with the split polygon
				ViewCell *viewCell = (coincident.size() > 0) ? 
					dynamic_cast<ViewCell *>(coincident.front()->mParent) : NULL;
				interior->ProcessPolygons(&coincident, mStorePolys);

				//Debug << "split node on level " << tData.mDepth << " bk: " << backPolys->size() << " frnt: " << frontPolys->size() << endl;
				mStat.splits += splits;

				// push the children on the stack
				tStack.push(BspTraversalData(interior->GetFront(), frontPolys, tData.mDepth + 1, NULL));
				tStack.push(BspTraversalData(interior->GetBack(), backPolys, tData.mDepth + 1, viewCell));
			}

			// cleanup
			DEL_PTR(tData.mPolygons);
		}
		else // reached leaf => subdivide current viewcell
		{
			if (tData.mPolygons->size() > 0)
				Debug << "Subdividing further: polygon size: " << (int)tData.mPolygons->size() << ", view cell: " << tData.mViewCell << endl;

			BspNode *root = Subdivide(tStack, tData, viewCells);		

			if (!mRoot) // tree empty => new root
				mRoot = root;
		}
	}
}

int BspTree::AddMesh2Polygons(Mesh *mesh, PolygonContainer &polys, Intersectable *parent)
{
	FaceContainer::const_iterator fi;
	int polysNum = 0;

	// copy the face data to polygons
	for (fi = mesh->mFaces.begin(); fi != mesh->mFaces.end(); ++ fi)
	{
		Polygon3 *poly = new Polygon3((*fi), mesh);
		poly->mParent = parent; // set parent intersectable
		polys.push_back(poly);

		if (!poly->CheckValid())
			Debug << "Input polygon not valid: " << *poly << endl;

		++ polysNum;
	}
	return polysNum;
}

int BspTree::Copy2PolygonSoup(const ViewCellContainer &viewCells, PolygonContainer &polys, int maxObjects)
{
	int limit = (maxObjects > 0) ? Min((int)viewCells.size(), maxObjects) : (int)viewCells.size();
  
	for (int i = 0; i < limit; ++i)
	{//if ((i == 12) ||(i==14))
		if (viewCells[i]->GetMesh()) // copy the mesh data to polygons
		{
			mBox.Include(viewCells[i]->GetBox()); // add to BSP tree aabb
			AddMesh2Polygons(viewCells[i]->GetMesh(), polys, viewCells[i]);
		}
	}

	return (int)polys.size();
}

int BspTree::Copy2PolygonSoup(const ObjectContainer &objects, PolygonContainer &polys, int maxObjects)
{
	int limit = (maxObjects > 0) ? Min((int)objects.size(), maxObjects) : (int)objects.size();
  
	for (int i = 0; i < limit; ++i)
	{
		Intersectable *object = objects[i];//*it;
		Mesh *mesh = NULL;

		switch (object->Type()) // extract the meshes
		{
		case Intersectable::MESH_INSTANCE:
			mesh = dynamic_cast<MeshInstance *>(object)->GetMesh();
			break;
		case Intersectable::VIEWCELL:
			mesh = dynamic_cast<ViewCell *>(object)->GetMesh();
			break;
			// TODO: handle transformed mesh instances
		default:
			break;
		}
		
        if (mesh) // copy the mesh data to polygons
		{
			mBox.Include(object->GetBox()); // add to BSP tree aabb
			AddMesh2Polygons(mesh, polys);
		}
	}

	return (int)polys.size();
}

void BspTree::Construct(const ViewCellContainer &viewCells)
{
	mStat.nodes = 1;
	mBox.Initialize(); 	// initialise bsp tree bounding box

 	// copy view cell meshes into one big polygon soup
	PolygonContainer *polys = new PolygonContainer();
	mStat.polys = Copy2PolygonSoup(viewCells, *polys);

	// construct tree from viewcell polygons
	Construct(polys);
	//Export("bsp.x3d");
}


void BspTree::Construct(const ObjectContainer &objects, ViewCellContainer *viewCells)
{
	mStat.nodes = 1;
	mBox.Initialize(); 	// initialise bsp tree bounding box
	
	PolygonContainer *polys = new PolygonContainer();
	
	// copy mesh instance polygons into one big polygon soup
	mStat.polys = Copy2PolygonSoup(objects, *polys);

	// construct tree from polygon soup
	Construct(polys, viewCells);
}

void BspTree::Construct(const RayContainer &rays, ViewCellContainer *viewCells)
{
	// TODO
}

void BspTree::Construct(PolygonContainer *polys, ViewCellContainer *viewCells)
{
	std::stack<BspTraversalData> tStack;
	
	BspTraversalData tData(new BspLeaf(), polys, 0, NULL);

	tStack.push(tData);

	long startTime = GetTime();
	Debug << "**** Contructing tree using objects ****\n";
	while (!tStack.empty()) 
	{
		tData = tStack.top();
	    tStack.pop();

		// subdivide leaf node
		BspNode *root = Subdivide(tStack, tData, viewCells);

		// empty tree => new root corresponding to unbounded space
		if (!mRoot) 
			mRoot = root;
	}

	Debug << "**** Finished tree construction ****\n";
	Debug << "BSP tree contruction time: " << TimeDiff(startTime, GetTime())*1e-3 << "s" << endl;
}


BspNode *BspTree::Subdivide(BspTraversalStack &tStack, BspTraversalData &tData, 
							ViewCellContainer *viewCells)
{
	//-- terminate traversal  
	if ((tData.mPolygons->size() <= sTermMaxPolygons) || (tData.mDepth >= sTermMaxDepth))
	{
#ifdef _DEBUG
		Debug << "subdivision terminated at depth " << tData.mDepth << ", #polys: " << (int)tData.mPolygons->size() << endl;
#endif

		EvaluateLeafStats(tData);

		BspLeaf *leaf = dynamic_cast<BspLeaf *>(tData.mNode);

		//-- new viewcells are generated and added to each leaf
		if (viewCells)
		{
			ViewCell *viewCell = new ViewCell();
			leaf->SetViewCell(viewCell);
			viewCells->push_back(viewCell);
			// Debug << "creating new viewcell" << endl;
		}
		//-- add viewcell stored in split polygon
		else if (tData.mViewCell)
		{
			if (leaf->GetViewCell())
				Debug << "ERROR: leaf already has view cell " << endl;//leaf->mViewCellIdx << endl;

			//leaf->mViewCellIdx = counter; Debug << "insert view cell " << tData.mViewCell << endl;
			leaf->SetViewCell(dynamic_cast<ViewCell *>(tData.mViewCell));
		}
	
		// remaining polygons are discarded or added to node
		tData.mNode->ProcessPolygons(tData.mPolygons, mStorePolys);
		DEL_PTR(tData.mPolygons);

		return tData.mNode;
	}

	//-- continue subdivision
	PolygonContainer *backPolys = new PolygonContainer();
	PolygonContainer *frontPolys = new PolygonContainer();
	PolygonContainer coincident;
	
	// create new interior node and two leaf nodes
	BspInterior *interior = SubdivideNode(dynamic_cast<BspLeaf *>(tData.mNode),
										  tData.mPolygons,
										  frontPolys,
										  backPolys, 
										  &coincident);

	// view cell associated with the first split polygon
	ViewCell *viewCell = (coincident.size() > 0) ? 
					dynamic_cast<ViewCell *>(coincident.front()->mParent) : NULL;
	interior->ProcessPolygons(&coincident, mStorePolys);

	// push the children on the stack
	// split polygon is only needed in the back leaf
	tStack.push(BspTraversalData(interior->GetFront(), frontPolys, tData.mDepth + 1, NULL));
	tStack.push(BspTraversalData(interior->GetBack(), backPolys, tData.mDepth + 1, viewCell));

	// cleanup
	DEL_PTR(tData.mNode);
	DEL_PTR(tData.mPolygons);

	return interior;
}


BspInterior *BspTree::SubdivideNode(BspLeaf *leaf, 
									PolygonContainer *polys, 
									PolygonContainer *frontPolys,
									PolygonContainer *backPolys, 
									PolygonContainer *coincident)
{
	mStat.nodes += 2;

	// add the new nodes to the tree + select subdivision plane
	BspInterior *interior = new BspInterior(SelectPlane(*polys)); 

#ifdef _DEBUG
	Debug << interior << endl;
#endif

	// split polygon according to current plane
	int splits = 0;
	
	interior->SplitPolygons(polys, frontPolys, backPolys, coincident, splits, mStorePolys);
	
 	mStat.splits += splits;

	BspInterior *parent = leaf->GetParent();

	// replace a link from node's parent
	if (parent)
	{
		parent->ReplaceChildLink(leaf, interior);
		interior->SetParent(parent);
	}

	// and setup child links
	interior->SetupChildLinks(new BspLeaf(interior), new BspLeaf(interior));
	
	return interior;
}

Plane3 BspTree::SelectPlane(const PolygonContainer &polys)  const
{
	if (polys.size() == 0)
	{
		Debug << "Warning: No split candidate available\n";
		return Plane3();
	}

	// simple strategy: just take next polygon
	if (sSplitPlaneStrategy & NEXT_POLYGON)
	{
		return polys.front()->GetSupportingPlane();
	}
	
	if (sSplitPlaneStrategy & AXIS_ALIGNED)
	{
		AxisAlignedBox3 box;
		box.Initialize();
	
		Polygon3::IncludeInBox(polys, box);
        
		int axis = box.Size().DrivingAxis();
		Vector3 norm(0,0,0); norm[axis] = 1;

		Vector3 pt = (box.Min()[axis] + box.Max()[axis]) * 0.5f;

		return Plane3(norm, pt);
	}

	/*if (sSplitPlaneStrategy & DRIVING_AXIS)
	{
		const float dot_thres = 0.999;
		PolygonContainer::const_iterator it, it_end = polys->end();
		
		for (it = polys->begin(); it != it_end; ++ it)
		{
			if (DotProd((*it)->GetSupportingPlane().mNormal, mVerticalAxis) > dot_thres)
				return (*it)->GetSupportingPlane();
		}
	}*/
	
	// use heuristics to find appropriate plane
	return SelectPlaneHeuristics(polys, sMaxCandidates);
}

Plane3 BspTree::SelectPlaneHeuristics(const PolygonContainer &polys, int maxTests) const
{
	float lowestCost = MAX_FLOAT;
	int bestPlaneIdx = 0;
	
	int limit = Min((int)polys.size(), maxTests);

	for (int i = 0; i < limit; ++ i)
	{
		int candidateIdx = GetNextCandidateIdx((int)polys.size());
		Plane3 candidatePlane = polys[candidateIdx]->GetSupportingPlane();
		
		// evaluate current candidate
		float candidateCost = EvalSplitPlane(polys, candidatePlane);
			
		if (candidateCost < lowestCost)
		{
			bestPlaneIdx = candidateIdx;
			lowestCost = candidateCost;
		}
	}
	//Debug << "Plane lowest cost: " << lowestCost << endl;
	
	return polys[bestPlaneIdx]->GetSupportingPlane();
}

int BspTree::GetNextCandidateIdx(const int max) const
{
	return Random(max); // TODO: avoid testing same plane 2 times
}

float BspTree::EvalSplitPlane(const PolygonContainer &polys, const Plane3 &candidatePlane) const
{
	float sumBalanced = 0, sumLeastSplits, drivingAxis = 0;
	
	if (sSplitPlaneStrategy & DRIVING_AXIS)
	{
        drivingAxis = (candidatePlane.mNormal.DrivingAxis() == mBox.Size().DrivingAxis()) ?
			99.0f : 0.0f; // always prefer driving axis => very high value
	}
	if ((sSplitPlaneStrategy & BALANCED_TREE) || (sSplitPlaneStrategy & LEAST_SPLITS))
	{
		PolygonContainer::const_iterator it, it_end = polys.end();
		
		for (it = polys.begin(); it != it_end; ++ it)
		{
			int classification = (*it)->ClassifyPlane(candidatePlane);

			if (sSplitPlaneStrategy & BALANCED_TREE)
				sumBalanced += sBalancedTreeTable[classification];
	
			if (sSplitPlaneStrategy & LEAST_SPLITS)
				sumLeastSplits += sLeastSplitsTable[classification];
		}
	}

	// return linear combination of the sums
	return fabs(sumBalanced) + sumLeastSplits + drivingAxis;
}

void BspTree::ParseEnvironment()
{
	environment->GetIntValue("BspTree.Termination.maxDepth", sTermMaxDepth);
	environment->GetIntValue("BspTree.Termination.maxPolygons", sTermMaxPolygons);
	environment->GetIntValue("BspTree.maxCandidates", sMaxCandidates);
	environment->GetIntValue("BspTree.splitPlaneStrategy", sSplitPlaneStrategy);

	Debug << "BSP max depth: " << sTermMaxDepth << endl;
	Debug << "BSP max polys: " << sTermMaxPolygons << endl;
	Debug << "BSP max candidates: " << sMaxCandidates << endl;

	//-- parse BSP tree construction method
	char constructionMethodStr[60];
	
	environment->GetStringValue("BspTree.constructionMethod", constructionMethodStr);

	sConstructionMethod = BspTree::VIEW_CELLS;
	
	if (strcmp(constructionMethodStr, "viewCells") == 0)
		sConstructionMethod = BspTree::VIEW_CELLS;
	else if (strcmp(constructionMethodStr, "sceneGeometry") == 0)
		sConstructionMethod = BspTree::SCENE_GEOMETRY;
	else if (strcmp(constructionMethodStr, "rays") == 0)
		sConstructionMethod = BspTree::RAYS;
	else 
	{
		cerr << "Wrong bsp construction method " << constructionMethodStr << endl;
		exit(1);
    }

	Debug << "Construction method: " << constructionMethodStr << endl;
}

void BspTree::CollectLeaves(vector<BspLeaf *> &leaves)
{
	stack<BspNode *> nodeStack;
	nodeStack.push(mRoot);
 
	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top();
    
		nodeStack.pop();
    
		if (node->IsLeaf()) 
		{
			BspLeaf *leaf = (BspLeaf *)node;
			
			leaves.push_back(leaf);
		} else 
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);
			nodeStack.push(interior->GetBack());
			nodeStack.push(interior->GetFront());
		}
	}
}

int BspTree::CollectLeafPvs()
{
	int totalPvsSize = 0;
  
	stack<BspNode *> nodeStack;
  
	nodeStack.push(mRoot);
  
	while (!nodeStack.empty()) 
	{
		BspNode *node = nodeStack.top();
		
		nodeStack.pop();
		
		if (node->IsLeaf()) 
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);

			ViewCell *viewcell = leaf->GetViewCell();
			
			if (!viewcell->Mailed()) 
			{
				viewcell->Mail(); // what does mail mean?
					
				// add this node to pvs of all nodes it can see
				BspPvsMap::iterator ni; 
	  
	/*			for (ni = object->mBspPvs.mEntries.begin(); ni != object->mKdPvs.mEntries.end(); ni++) 
				{
					BspNode *node = (*ni).first;
	    
					// $$ JB TEMPORARY solution -> should add object PVS or explictly computed
					// BSP tree PVS
	    			if (leaf->mBspPvs.AddNodeSample(node))
						totalPvsSize++;
				}*/
			}
		} else 
		{
			// traverse tree
			BspInterior *interior = (BspInterior *)node;
      
			nodeStack.push(interior->GetFront());
			nodeStack.push(interior->GetBack());
		}
	}

	return totalPvsSize;
}

AxisAlignedBox3 BspTree::GetBoundingBox() const
{
	return mBox;
}

BspNode *BspTree::GetRoot() const
{
	return mRoot;
}

bool BspTree::StorePolys() const
{
	return mStorePolys;
}

void BspTree::EvaluateLeafStats(const BspTraversalData &data)
{
	// the node became a leaf -> evaluate stats for leafs
	BspLeaf *leaf = dynamic_cast<BspLeaf *>(data.mNode);

	if (data.mDepth >= sTermMaxDepth)
	{
		++ mStat.maxDepthNodes;
	}

	// store maximal and minimal depth
	if (data.mDepth > mStat.maxDepth)
		mStat.maxDepth = data.mDepth;
	if (data.mDepth < mStat.minDepth)
		mStat.minDepth = data.mDepth;

	// accumulate depth to compute average
	mStat.accumDepth += data.mDepth;

	if (leaf->mViewCell)
		++ mStat.viewCellLeaves;

#ifdef _DEBUG
	Debug << "BSP Traversal data. Depth: " << data.mDepth << " (max: " << mTermMaxDepth << "), #polygons: " << 
	  data.mPolygons->size() << " (max: " << sTermMaxPolygons << ")" << endl;
#endif
}

int BspTree::CastRay(Ray &ray)
{
	int hits = 0;
  
	stack<BspRayTraversalData> tStack;
  
	float maxt = 1e6;
	float mint = 0;

	Intersectable::NewMail();

	// test with tree bounding box
	if (!mBox.GetMinMaxT(ray, &mint, &maxt))
		return 0;

	if (mint < 0)
		mint = 0;
  
	maxt += Limits::Threshold;
  
	Vector3 entp = ray.Extrap(mint);
	Vector3 extp = ray.Extrap(maxt);
  
	BspNode *node = mRoot;
	BspNode *farChild;
	
	while (1)
	{
		if (!node->IsLeaf()) 
		{
			BspInterior *in = (BspInterior *) node;
			
			Plane3 *splitPlane = in->GetPlane();

			float t = 0;
			bool coplanar = false;
		
			int entSide = splitPlane->Side(entp);
			int extSide = splitPlane->Side(extp);

			Vector3 intersection;

			if (entSide < 0)
			{
				node = in->GetBack();

				if(extSide <= 0) 
					continue;
					
				farChild = in->GetFront(); // plane splits ray

			} else if (entSide > 0)
			{
				node = in->GetFront();

				if (extSide >= 0)
					continue;

				farChild = in->GetBack();	 // plane splits ray			
			}
			else // ray and plane are coincident // WHAT TO DO IN THIS CASE ?
			{
				node = in->GetFront();
				continue;
			}

			//-- split

			// push data for far child
			tStack.push(BspRayTraversalData(farChild, extp, maxt));

			// find intersection with plane between ray origin and exit point
			extp = splitPlane->FindIntersection(ray.GetLoc(), extp, &maxt);

		} else // compute intersections with objects in leaf
		{
			BspLeaf *leaf = dynamic_cast<BspLeaf *>(node);
			//ray.leaves.push_back(leaf);
      // TODO
   			/*ObjectContainer::const_iterator mi;
			for (mi = leaf->mObjects.begin(); mi != leaf->mObjects.end(); ++mi) 
			{
				Intersectable *object = *mi;
				if (!object->Mailed()) 
				{
					object->Mail();
					//ray.meshes.push_back(mesh);
					hits += object->CastRay(ray);
				}
			}*/

			if (hits && ray.GetType() == Ray::LOCAL_RAY)
				if (ray.intersections[0].mT <= maxt)
					break;
      
			// get the next node from the stack
			if (tStack.empty())
				break;
      
			entp = extp;
			mint = maxt;
			BspRayTraversalData &s  = tStack.top();

			node = s.mNode;
			extp = s.mExitPoint;
			maxt = s.mMaxT;

			tStack.pop();
		}
	}

	return hits;
}

bool BspTree::Export(const string filename)
{
	Exporter *exporter = Exporter::GetExporter(filename);

	if (exporter) 
	{
		exporter->ExportBspTree(*this);
		return true;
	}	

	return false;
}
//} // GtpVisibilityPreprocessor
